"""Sphinx configuration of SHRY documentation."""

# -*- coding: utf-8 -*-
#
# SHRY documentation build configuration file, created by
# sphinx-quickstart on Mon Apr 13 15:11:21 2009.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# import sphinx_bootstrap_theme

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.append(os.path.abspath('.'))

import os
import sys
import subprocess
from datetime import datetime

# Add the project root to sys.path so that we can import the package
sys.path.insert(0, os.path.abspath(".."))

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.mathjax",
    "myst_parser",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinxarg.ext",
]
myst_enable_extensions = ["linkify", "dollarmath", "amsmath"]

# Napoleon: parse Google/NumPy style docstrings (used in this project) and
# render attribute/parameter tables in autodoc pages.
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_ivar = True
napoleon_attr_annotations = True
napoleon_preprocess_types = True

# Move type hints from signatures into the description so Attributes/Parameters
# tables remain readable and avoid duplicated type-only attribute listings.
autodoc_typehints = "description"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = "utf-8"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "SHRY"
author = "Kosuke Nakano, Genki Imam Prayogo"
copyright = f"{datetime.now().year}, {author}"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
try:
    from importlib import metadata
    import re

    raw_version = metadata.version("shry")
    # Remove .dev..., +..., and everything after
    # Example: 0.1.dev28+gxxx -> 0.1
    # Example: 0.1.0a1.dev5 -> 0.1.0a1
    release = re.sub(r"(\.dev\d+|[+-].*)", "", raw_version)
    version = ".".join(release.split(".")[:2])
except Exception:
    release = "0.0.0"
    version = "0.0"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = ["_build"]
exclude_patterns = ["README.md"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme = 'default'
# html_theme = 'nature'
# html_theme = "bootstrap"
# html_theme = 'sphinxbootstrap4theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "icon_links": [
        {"name": "GitHub", "url": "https://github.com/shry-project/SHRY", "icon": "fa-brands fa-github"},
    ],
}
# html_theme_options = {
#     # Navigation bar title. (Default: ``project`` value)
#     "navbar_title": "SHRY",
#     # Tab name for entire site. (Default: "Site")
#     "navbar_site_name": "Site",
#     # A list of tuples containing pages or urls to link to.
#     # Valid tuples should be in the following forms:
#     #    (name, page)                 # a link to a page
#     #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
#     #    (name, "http://example.com", True) # arbitrary absolute url
#     # Note the "1" or "True" value above as the third argument to indicate
#     # an arbitrary url.
#     # 'navbar_links': [
#     #     ("Examples", "examples"),
#     #     ("Link", "http://example.com", True),
#     #     ],
#     # 'navbar_links': [
#     #     ("Tags", "setting-tags"),
#     #     ("Options", "command-options"),
#     #     ],
#     # Render the next and previous page links in navbar. (Default: true)
#     "navbar_sidebarrel": True,
#     # Render the current pages TOC in the navbar. (Default: true)
#     "navbar_pagenav": True,
#     # Global TOC depth for "site" navbar tab. (Default: 1)
#     # Switching to -1 shows all levels.
#     "globaltoc_depth": 1,
#     # Include hidden TOCs in Site navbar?
#     #
#     # Note: If this is "false", you cannot have mixed ``:hidden:`` and
#     # non-hidden ``toctree`` directives in the same page, or else the build
#     # will break.
#     #
#     # Values: "true" (default) or "false"
#     "globaltoc_includehidden": "true",
#     # HTML navbar class (Default: "navbar") to attach to <div> element.
#     # For black navbar, do "navbar navbar-inverse"
#     # 'navbar_class': "navbar navbar-inverse",
#     "navbar_class": "navbar",
#     # Fix navigation bar to top of page?
#     # Values: "true" (default) or "false"
#     "navbar_fixed_top": "true",
#     # Location of link to source.
#     # Options are "nav" (default), "footer" or anything else to exclude.
#     # 'source_link_position': "nav",
#     "source_link_position": "footer",
#     # Bootswatch (http://bootswatch.com/) theme.
#     #
#     # Options are nothing with "" (default) or the name of a valid theme
#     # such as "amelia" or "cosmo".
#     # 'bootswatch_theme': "united",
#     "bootswatch_theme": "cerulean",
#     # Choose Bootstrap version.
#     # Values: "3" (default) or "2" (in quotes)
#     "bootstrap_version": "3",
# }

html_theme = "sphinx_book_theme"

#
# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []
# html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()
# html_theme_path = [sphinxbootstrap4theme.get_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "SHRY v%s" % release

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sideba.
# html_logo = '_static/logo_yoko2.jpg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "jQMCdoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("documentation", "jQMC.tex", "jQMC manual", "Kosuke Nakano", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True


def run_scripts(app):
    """Run scripts to generate images."""
    script_path = os.path.join(app.srcdir, "notes/paulinet_rbf_plot.py")
    if os.path.exists(script_path):
        print(f"Running {script_path}...")
        subprocess.call(["python", script_path], cwd=os.path.dirname(script_path))


def setup(app):
    app.connect("builder-inited", run_scripts)
